﻿/*
  This file is part of Sanmill.
  Copyright (C) 2019-2021 The Sanmill developers (see AUTHORS file)

  Sanmill is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Sanmill is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iomanip>
#include <sstream>

#include "bitboard.h"
#include "position.h"
#include "thread.h"
#include "mills.h"
#include "option.h"

using std::string;

namespace Zobrist
{
constexpr int KEY_MISC_BIT = 2;
Key psq[PIECE_TYPE_NB][SQUARE_NB];
Key side;
}

namespace
{
const string  PieceToChar(Piece p)
{
    if (p == NO_PIECE) {
        return "*";
    }

    if (p == BAN_STONE) {
        return "X";
    }

    if (W_STONE <= p && p <= W_STONE_12) {
        return "O";
    }

    if (B_STONE <= p && p <= B_STONE_12) {
        return "@";
    }

    return "*";
}

Piece CharToPiece(char ch) noexcept
{
    if (ch == '*') {
        return NO_PIECE;
    }

    if (ch == 'O') {
        return W_STONE;
    }

    if (ch == '@') {
        return B_STONE;
    }

    if (ch == 'X') {
        return BAN_STONE;
    }

    return NO_PIECE;
}

constexpr PieceType PieceTypes[] = { NO_PIECE_TYPE, WHITE_STONE, BLACK_STONE, BAN };
} // namespace


/// operator<<(Position) returns an ASCII representation of the position

std::ostream &operator<<(std::ostream &os, const Position &pos)
{
    /*
        X --- X --- X
        |\    |    /|
        | X - X - X |
        | |\  |  /| |
        | | X-X-X | |
        X-X-X   X-X-X
        | | X-X-X | |
        | |/  |  \| |
        | X - X - X |
        |/    |    \|
        X --- X --- X
    */

    /*
        31 ----- 24 ----- 25
        | \       |      / |
        |  23 -- 16 -- 17  |
        |  | \    |   / |  |
        |  |  15 08 09  |  |
        30-22-14    10-18-26
        |  |  13 12 11  |  |
        |  | /    |   \ |  |
        |  21 -- 20 -- 19  |
        | /       |     \  |
        29 ----- 28 ----- 27
    */

#define P(s) PieceToChar(pos.piece_on(Square(s)))

    if (rule.hasDiagonalLines) {
        os << "\n";
        os << P(31) << " --- " << P(24) << " --- " << P(25) << "\n";
        os << "|\\    |    /|\n";
        os << "| " << P(23) << " - " << P(16) << " - " << P(17) << " |\n";
        os << "| |\\  |  /| |\n";
        os << "| | " << P(15) << "-" << P(8) << "-" << P(9) << " | |\n";
        os << P(30) << "-" << P(22) << "-" << P(14) << "   " << P(10) << "-" << P(18) << "-" << P(26) << "\n";
        os << "| | " << P(13) << "-" << P(12) << "-" << P(11) << " | |\n";
        os << "| |/  |  \\| |\n";
        os << "| " << P(21) << " - " << P(20) << " - " << P(19) << " |\n";
        os << "|/    |    \\|\n";
        os << P(29) << " --- " << P(28) << " --- " << P(27) << "\n";
    } else {
        os << "\n";
        os << P(31) << " --- " << P(24) << " --- " << P(25) << "\n";
        os << "|     |     |\n";
        os << "| " << P(23) << " - " << P(16) << " - " << P(17) << " |\n";
        os << "| |   |   | |\n";
        os << "| | " << P(15) << "-" << P(8) << "-" << P(9) << " | |\n";
        os << P(30) << "-" << P(22) << "-" << P(14) << "   " << P(10) << "-" << P(18) << "-" << P(26) << "\n";
        os << "| | " << P(13) << "-" << P(12) << "-" << P(11) << " | |\n";
        os << "| |   |   | |\n";
        os << "| " << P(21) << " - " << P(20) << " - " << P(19) << " |\n";
        os << "|     |     |\n";
        os << P(29) << " --- " << P(28) << " --- " << P(27) << "\n";
    }

#undef P

    os << "\nFen: " << pos.fen() << "\nKey: " << std::hex << std::uppercase
        << std::setfill('0') << std::setw(16) << pos.key();

    return os;
}


/// Position::init() initializes at startup the various arrays used to compute hash keys

void Position::init()
{
    PRNG rng(1070372);

    for (PieceType pt : PieceTypes)
        for (Square s = SQ_BEGIN; s < SQ_END; ++s)
            Zobrist::psq[pt][s] = rng.rand<Key>() << Zobrist::KEY_MISC_BIT >> Zobrist::KEY_MISC_BIT;

    Zobrist::side = rng.rand<Key>() << Zobrist::KEY_MISC_BIT >> Zobrist::KEY_MISC_BIT;

    return;
}

Position::Position()
{
    construct_key();

    reset();

    score[WHITE] = score[BLACK] = score_draw = gamesPlayedCount = 0;
}


/// Position::set() initializes the position object with the given FEN string.
/// This function is not very robust - make sure that input FENs are correct,
/// this is assumed to be the responsibility of the GUI.

Position &Position::set(const string &fenStr, Thread *th)
{
    /*
       A FEN string defines a particular position using only the ASCII character set.

       A FEN string contains six fields separated by a space. The fields are:

       1) Piece placement. Each rank is described, starting
          with rank 1 and ending with rank 8. Within each rank, the contents of each
          square are described from file A through file C. Following the Standard
          Algebraic Notation (SAN), each piece is identified by a single letter taken
          from the standard English names. White pieces are designated using "O"
          whilst Black uses "@". Blank uses "*". Banned uses "X".
          noted using digits 1 through 8 (the number of blank squares), and "/"
          separates ranks.

       2) Active color. "w" means white moves next, "b" means black.

       3) Phrase.

       4) Action.

       5) White on board/White in hand/Black on board/Black in hand/need to remove

       6) Halfmove clock. This is the number of halfmoves since the last
          capture. This is used to determine if a draw can be claimed under the
          fifty-move rule.

       7) Fullmove number. The number of the full move. It starts at 1, and is
          incremented after White's move.
    */

    unsigned char token = '\0';
    Square sq = SQ_A1;
    std::istringstream ss(fenStr);

    std::memset(this, 0, sizeof(Position));

    ss >> std::noskipws;

    // 1. Piece placement
    while ((ss >> token) && !isspace(token)) {
        if (token == 'O' || token == '@' || token == 'X') {
            put_piece(CharToPiece(token), sq);
            ++sq;
        }
        if (token == '*') {
            ++sq;
        }
    }

    // 2. Active color
    ss >> token;
    sideToMove = (token == 'w' ? WHITE : BLACK);
    them = ~sideToMove;    // Note: Stockfish do not need to set them

    // 3. Phrase
    ss >> token;
    ss >> token;

    switch (token) {
    case 'r':
        phase = Phase::ready;
        break;
    case 'p':
        phase = Phase::placing;
        break;
    case 'm':
        phase = Phase::moving;
        break;
    case 'o':
        phase = Phase::gameOver;
        break;
    default:
        phase = Phase::none;
    }

    // 4. Action
    ss >> token;
    ss >> token;

    switch (token) {
    case 'p':
        action = Action::place;
        break;
    case 's':
        action = Action::select;
        break;
    case 'r':
        action = Action::remove;
        break;
    default:
        action = Action::none;
    }

    // 5. White on board / White in hand / Black on board / Black in hand / need to remove
    ss >> std::skipws
        >> pieceOnBoardCount[WHITE] >> pieceInHandCount[WHITE]
        >> pieceOnBoardCount[BLACK] >> pieceInHandCount[BLACK]
        >> pieceToRemoveCount;


    // 6-7. Halfmove clock and fullmove number
    ss >> std::skipws >> st.rule50 >> gamePly;

    // Convert from fullmove starting from 1 to gamePly starting from 0,
    // handle also common incorrect FEN with fullmove = 0.
    gamePly = std::max(2 * (gamePly - 1), 0) + (sideToMove == BLACK);

    thisThread = th;

    return *this;
}


/// Position::fen() returns a FEN representation of the position.
/// This is mainly a debugging function.

const string Position::fen() const
{
    std::ostringstream ss;

    // Piece placement data
    for (File f = FILE_A; f <= FILE_C; ++f) {
        for (Rank r = RANK_1; r <= RANK_8; ++r) {
            ss << PieceToChar(piece_on(make_square(f, r)));
        }

        if (f == FILE_C) {
            ss << " ";
        } else {
            ss << "/";
        }
    }

    // Active color
    ss << (sideToMove == WHITE ? "w" : "b");

    ss << " ";

    // Phrase
    switch (phase) {
    case Phase::none:
        ss << "n";
        break;
    case Phase::ready:
        ss << "r";
        break;
    case Phase::placing:
        ss << "p";
        break;
    case Phase::moving:
        ss << "m";
        break;
    case Phase::gameOver:
        ss << "o";
        break;
    default:
        ss << "?";
        break;
    }

    ss << " ";

    // Action
    switch (action) {
    case Action::place:
        ss << "p";
        break;
    case Action::select:
        ss << "s";
        break;
    case Action::remove:
        ss << "r";
        break;
    default:
        ss << "?";
        break;
    }

    ss << " ";

    ss << pieceOnBoardCount[WHITE] << " " << pieceInHandCount[WHITE] << " "
        << pieceOnBoardCount[BLACK] << " " << pieceInHandCount[BLACK] << " "
        << pieceToRemoveCount << " ";

    ss << st.rule50 << " " << 1 + (gamePly - (sideToMove == BLACK)) / 2;

    return ss.str();
}


/// Position::legal() tests whether a pseudo-legal move is legal

bool Position::legal(Move m) const
{
    assert(is_ok(m));

    const Color us = sideToMove;
    const Square from = from_sq(m);
    const Square to = to_sq(m);

    if (from == to) {
        return false;
    }

    if (phase == Phase::moving && type_of(move) != MOVETYPE_REMOVE) {
        if (color_of(moved_piece(m)) != us) {
            return false;
        }
    }

    return true;
}


/// Position::do_move() makes a move, and saves all information necessary
/// to a StateInfo object. The move is assumed to be legal. Pseudo-legal
/// moves should be filtered out before this function is called.

void Position::do_move(Move m)
{
    bool ret = false;

    const MoveType mt = type_of(m);

    switch (mt) {
    case MOVETYPE_REMOVE:
        ret = remove_piece(to_sq(m));
        if (ret) {
            // Reset rule 50 counter
            st.rule50 = 0;
        }
        break;
    case MOVETYPE_MOVE:
        ret = move_piece(from_sq(m), to_sq(m));
        if (ret) {
            ++st.rule50;
        }
        break;
    case MOVETYPE_PLACE:
        ret = put_piece(to_sq(m));
        if (ret) {
            // Reset rule 50 counter
            st.rule50 = 0;
        }
        break;
    default:
        break;
    }

    if (!ret) {
        return;
    }

    // Increment ply counters. In particular
    ++gamePly;
    ++st.pliesFromNull;

    move = m;
}


/// Position::undo_move() unmakes a move. When it returns, the position should
/// be restored to exactly the same state as before the move was made.

void Position::undo_move(Sanmill::Stack<Position> &ss)
{
    memcpy(this, ss.top(), sizeof(Position));
    ss.pop();
}


/// Position::key_after() computes the new hash key after the given move. Needed
/// for speculative prefetch. It doesn't recognize special moves like (need remove)

Key Position::key_after(Move m) const
{
    Key k = st.key;
    const Square s = static_cast<Square>(to_sq(m));;
    const MoveType mt = type_of(m);

    if (mt == MOVETYPE_REMOVE) {
        k ^= Zobrist::psq[~side_to_move()][s];

        if (rule.hasBannedLocations && phase == Phase::placing) {
            k ^= Zobrist::psq[BAN][s];
        }

        goto out;
    }

    k ^= Zobrist::psq[side_to_move()][s];

    if (mt == MOVETYPE_MOVE) {
        k ^= Zobrist::psq[side_to_move()][from_sq(m)];
    }

out:
    k ^= Zobrist::side;

    return k;
}

int repetition;

#ifdef THREEFOLD_REPETITION
// Position::has_repeated() tests whether there has been at least one repetition
// of positions since the last remove.

bool Position::has_repeated(Sanmill::Stack<Position> &ss) const
{
    for (int i = (int)posKeyHistory.size() - 2; i >= 0; i--) {
        if (key() == posKeyHistory[i]) {
            return true;
        }
    }

    int size = ss.size();

    for (int i = size - 1; i >= 0; i--) {
        if (type_of(ss[i].move) == MOVETYPE_REMOVE) {
            break;
        }
        if (key() == ss[i].st.key) {
            return true;
        }
    }

    return false;
}


/// Position::has_game_cycle() tests if the position has a move which draws by repetition.

bool Position::has_game_cycle() const
{
    for (auto i : posKeyHistory) {
        if (key() == i) {
            repetition++;
            if (repetition == 3) {
                repetition = 0;
                return true;
            }
        }
    }

    return false;
}
#endif // THREEFOLD_REPETITION


/// Mill Game

extern int repetition;

bool Position::reset()
{
    repetition = 0;

    gamePly = 0;
    st.rule50 = 0;

    phase = Phase::ready;
    set_side_to_move(WHITE);
    action = Action::place;

    winner = NOBODY;
    gameOverReason = GameOverReason::noReason;

    memset(board, 0, sizeof(board));
    memset(byTypeBB, 0, sizeof(byTypeBB));
    memset(byColorBB, 0, sizeof(byColorBB));

    st.key = 0;

    pieceOnBoardCount[WHITE] = pieceOnBoardCount[BLACK] = 0;
    pieceInHandCount[WHITE] = pieceInHandCount[BLACK] = rule.piecesCount;
    pieceToRemoveCount = 0;

    mobilityDiff = 0;

    MoveList<LEGAL>::create();
    create_mill_table();
    currentSquare = SQ_0;

#ifdef ENDGAME_LEARNING
    if (gameOptions.isEndgameLearningEnabled() &&
        gamesPlayedCount > 0 && gamesPlayedCount % SAVE_ENDGAME_EVERY_N_GAMES == 0) {
        Thread::saveEndgameHashMapToFile();
    }
#endif /* ENDGAME_LEARNING */

    int r;
    for (r = 0; r < N_RULES; r++) {
        if (strcmp(rule.name, RULES[r].name) == 0)
            break;
    }

    if (snprintf(record, RECORD_LEN_MAX, "r%1d s%03zu t%02d",
                r + 1, rule.maxStepsLedToDraw, 0) > 0) {
        return true;
    }

    record[0] = '\0';

    return false;
}

bool Position::start()
{
    gameOverReason = GameOverReason::noReason;

    switch (phase) {
    case Phase::placing:
    case Phase::moving:
        return false;
    case Phase::gameOver:
        reset();
        [[fallthrough]];
    case Phase::ready:
        phase = Phase::placing;
        return true;
    default:
        return false;
    }
}

bool Position::put_piece(Square s, bool updateRecord)
{
    Piece piece = NO_PIECE;
    Color us = sideToMove;

    if (phase == Phase::gameOver ||
        action != Action::place ||
        !(SQ_BEGIN <= s && s < SQ_END) || board[s]) {
        return false;
    }

    if (phase == Phase::ready) {
        start();
    }

    if (phase == Phase::placing) {
        piece = (Piece)((0x01 | make_piece(sideToMove)) + rule.piecesCount - pieceInHandCount[us]);
        pieceInHandCount[us]--;
        pieceOnBoardCount[us]++;

        const Piece pc = board[s] = piece;
        byTypeBB[ALL_PIECES] |= byTypeBB[type_of(pc)] |= s;
        byColorBB[color_of(pc)] |= s;   // TODO: Put ban?

        update_key(s);

        updateMobility(MOVETYPE_PLACE, s);

        if (updateRecord) {
            snprintf(record, RECORD_LEN_MAX, "(%1u,%1u)", file_of(s), rank_of(s));
        }

        currentSquare = s;

#ifdef MUEHLE_NMM
        if (pieceInHandCount[WHITE] == 0 &&
            pieceInHandCount[BLACK] == 0 &&
            is_all_surrounded(~sideToMove, SQ_0, s)) {
            set_gameover(sideToMove, GameOverReason::loseReasonNoWay);
            //change_side_to_move();
            return true;
        }
#endif

        const int n = mills_count(currentSquare);

        if (n == 0
#ifdef MUEHLE_NMM
            || is_all_in_mills(them)
#endif
            ) {
            assert(pieceInHandCount[WHITE] >= 0 && pieceInHandCount[BLACK] >= 0);

            if (pieceInHandCount[WHITE] == 0 && pieceInHandCount[BLACK] == 0) {
                if (check_if_game_is_over()) {
                    return true;
                }

                phase = Phase::moving;
                action = Action::select;

                if (rule.hasBannedLocations) {
                    remove_ban_stones();
                }

                if (!rule.isDefenderMoveFirst) {
                    change_side_to_move();
                }

                if (check_if_game_is_over()) {
                    return true;
                }
            } else {
                change_side_to_move();
            }
        } else {
            pieceToRemoveCount = rule.mayRemoveMultiple ? n : 1;
            update_key_misc();
            action = Action::remove;
        }

    } else if (phase == Phase::moving) {

#ifdef MUEHLE_NMM
        if (is_all_surrounded(~sideToMove, currentSquare, s)) {
            set_gameover(sideToMove, GameOverReason::loseReasonNoWay);
        }
#else
        if (check_if_game_is_over()) {
            return true;
        }
#endif // MUEHLE_NMM

        // If illegal
        if (pieceOnBoardCount[sideToMove] > rule.flyPieceCount ||
            !rule.mayFly) {
            if ((square_bb(s) & MoveList<LEGAL>::adjacentSquaresBB[currentSquare]) == 0) {
                return false;
            }
        }

        if (updateRecord) {
            snprintf(record, RECORD_LEN_MAX, "(%1u,%1u)->(%1u,%1u)",
                    file_of(currentSquare), rank_of(currentSquare),
                    file_of(s), rank_of(s));
            st.rule50++;
        }

        const Piece pc = board[currentSquare];

        CLEAR_BIT(byTypeBB[ALL_PIECES], currentSquare);
        CLEAR_BIT(byTypeBB[type_of(pc)], currentSquare);
        CLEAR_BIT(byColorBB[color_of(pc)], currentSquare);

        updateMobility(MOVETYPE_REMOVE, currentSquare);

        SET_BIT(byTypeBB[ALL_PIECES], s);
        SET_BIT(byTypeBB[type_of(pc)], s);
        SET_BIT(byColorBB[color_of(pc)], s);

        updateMobility(MOVETYPE_PLACE, s);

        board[s] = pc;
        update_key(s);
        revert_key(currentSquare);

        board[currentSquare] = NO_PIECE;

        currentSquare = s;
        
        const int n = mills_count(currentSquare);

        if (n == 0
#ifdef MUEHLE_NMM
            || is_all_in_mills(them)
#endif
            ) {
            action = Action::select;
            change_side_to_move();

            if (check_if_game_is_over()) {
                return true;
            }
        } else {
            pieceToRemoveCount = rule.mayRemoveMultiple ? n : 1;
            update_key_misc();
            action = Action::remove;
        }
    } else {
        assert(0);
    }

    return true;
}

bool Position::remove_piece(Square s, bool updateRecord)
{
    if (phase == Phase::ready || phase == Phase::gameOver)
        return false;

    if (action != Action::remove)
        return false;

    if (pieceToRemoveCount <= 0)
        return false;

    // if piece is not their
    if (!(make_piece(~side_to_move()) & board[s]))
        return false;

    if (!rule.mayRemoveFromMillsAlways &&
        potential_mills_count(s, NOBODY)
#ifndef MUEHLE_NMM
        && !is_all_in_mills(~sideToMove)
#endif
       ) {
        return false;
    }

    revert_key(s);

    if (rule.hasBannedLocations && phase == Phase::placing) {
        // Remove and put ban
        Piece pc = board[s];

        CLEAR_BIT(byTypeBB[type_of(pc)], s);    // TODO
        CLEAR_BIT(byColorBB[color_of(pc)], s);

        updateMobility(MOVETYPE_REMOVE, s);

        pc = board[s] = BAN_STONE;
        update_key(s);

        SET_BIT(byTypeBB[type_of(pc)], s);
    } else {
        // Remove only
        const Piece pc = board[s];

        CLEAR_BIT(byTypeBB[ALL_PIECES], s);
        CLEAR_BIT(byTypeBB[type_of(pc)], s);
        CLEAR_BIT(byColorBB[color_of(pc)], s);

        updateMobility(MOVETYPE_REMOVE, s);

        board[s] = NO_PIECE;
    }

    if (updateRecord) {
        snprintf(record, RECORD_LEN_MAX, "-(%1u,%1u)", file_of(s), rank_of(s));
        st.rule50 = 0;     // TODO: Need to move out?
    }

    pieceOnBoardCount[them]--;

    if (pieceOnBoardCount[them] + pieceInHandCount[them] < rule.piecesAtLeastCount) {
        set_gameover(sideToMove, GameOverReason::loseReasonlessThanThree);
        return true;
    }

    currentSquare = SQ_0;

    pieceToRemoveCount--;
    update_key_misc();

    if (pieceToRemoveCount > 0) {
        return true;
    }

    if (phase == Phase::placing) {
        if (pieceInHandCount[WHITE] == 0 && pieceInHandCount[BLACK] == 0) {
            phase = Phase::moving;
            action = Action::select;

            if (rule.hasBannedLocations) {
                remove_ban_stones();
            }

            if (rule.isDefenderMoveFirst) {
                goto check;
            }
        } else {
            action = Action::place;
        }
    } else {
        action = Action::select;
    }

    change_side_to_move();

check:
    check_if_game_is_over();

    return true;
}

bool Position::select_piece(Square s)
{
    if (phase != Phase::moving)
        return false;

    if (action != Action::select && action != Action::place)
        return false;

    if (board[s] & make_piece(sideToMove)) {
        currentSquare = s;
        action = Action::place;

        return true;
    }

    return false;
}

bool Position::resign(Color loser)
{
    if (phase == Phase::ready ||
        phase == Phase::gameOver ||
        phase == Phase::none) {
        return false;
    }

    set_gameover(~loser, GameOverReason::loseReasonResign);

    snprintf(record, RECORD_LEN_MAX, loseReasonResignStr, loser);

    return true;
}

bool Position::command(const char *cmd)
{
    unsigned int ruleNo = 0;
    unsigned t = 0;
    int step = 0;
    File file1 = FILE_A, file2 = FILE_A;
    Rank rank1 = RANK_1, rank2 = RANK_1;
    int args = 0;

    if (sscanf(cmd, "r%1u s%3d t%2u", &ruleNo, &step, &t) == 3) {
        if (set_rule(ruleNo - 1) == false) {
            return false;
        }

        return reset();
    }

    args = sscanf(cmd, "(%1u,%1u)->(%1u,%1u)", (unsigned*)&file1, (unsigned*)&rank1, (unsigned*)&file2, (unsigned*)&rank2);

    if (args >= 4) {
        return move_piece(file1, rank1, file2, rank2);
    }

    args = sscanf(cmd, "-(%1u,%1u)", (unsigned *)&file1, (unsigned *)&rank1);
    if (args >= 2) {
        return remove_piece(file1, rank1);
    }

    args = sscanf(cmd, "(%1u,%1u)", (unsigned *)&file1, (unsigned *)&rank1);
    if (args >= 2) {
        return put_piece(file1, rank1);
    }

    args = sscanf(cmd, "Player%1u give up!", &t);

    if (args == 1) {
        return resign((Color)t);
    }

#ifdef THREEFOLD_REPETITION
    if (!strcmp(cmd, drawReasonThreefoldRepetitionStr)) {
        return true;
    }

    if (!strcmp(cmd, "draw")) {
        set_gameover(DRAW, GameOverReason::drawReasonThreefoldRepetition);
        //snprintf(record, RECORD_LEN_MAX, drawReasonThreefoldRepetitionStr);
        return true;
    }
#endif /* THREEFOLD_REPETITION */

    return false;
}

Color Position::get_winner() const noexcept
{
    return winner;
}

void Position::set_gameover(Color w, GameOverReason reason)
{
    phase = Phase::gameOver;
    gameOverReason = reason;
    winner = w;

    update_score();
}

void Position::update_score()
{
    if (phase == Phase::gameOver) {
        if (winner == DRAW) {
            score_draw++;
            return;
        }

        score[winner]++;
    }
}

bool Position::check_if_game_is_over()
{
    if (phase == Phase::ready ||
        phase == Phase::gameOver) {
        return true;
    }

#ifdef RULE_50
    if (rule.maxStepsLedToDraw > 0 &&
        posKeyHistory.size() > rule.maxStepsLedToDraw) {
        set_gameover(DRAW, GameOverReason::drawReasonRule50);
        return true;
    }
#endif // RULE_50

    if (pieceOnBoardCount[WHITE] + pieceOnBoardCount[BLACK] >= EFFECTIVE_SQUARE_NB) {
        if (rule.isWhiteLoseButNotDrawWhenBoardFull) {
            set_gameover(BLACK, GameOverReason::loseReasonBoardIsFull);
        } else {
            set_gameover(DRAW, GameOverReason::drawReasonBoardIsFull);
        }

        return true;
    }

    if (phase == Phase::moving && action == Action::select && is_all_surrounded(sideToMove)) {
        if (rule.isLoseButNotChangeSideWhenNoWay) {
            set_gameover(~sideToMove, GameOverReason::loseReasonNoWay);
            return true;
        } else {
            change_side_to_move();  // TODO: Need?
            return false;
        }
    }

    return false;
}

int Position::calculate_mobility_diff()
{
    // TODO: Deal with rule is no ban location
    int mobilityWhite = 0;
    int mobilityBlack = 0;

    for (Square s = SQ_BEGIN; s < SQ_END; ++s) {
        if (board[s] == NO_PIECE || board[s] == BAN_STONE) {
            Square moveSquare;
            for (MoveDirection d = MD_BEGIN; d < MD_NB; ++d) {
                moveSquare = static_cast<Square>(MoveList<LEGAL>::adjacentSquares[s][d]);
                if (moveSquare) {
                    if (board[moveSquare] & W_STONE) {
                        mobilityWhite++;
                    }
                    if (board[moveSquare] & B_STONE) {
                        mobilityBlack++;
                    }
                }
            }
        }
    }

    return mobilityWhite - mobilityBlack;
}

void Position::remove_ban_stones()
{
    assert(rule.hasBannedLocations);

    Square s = SQ_0;

    for (int f = 1; f <= FILE_NB; f++) {
        for (int r = 0; r < RANK_NB; r++) {
            s = static_cast<Square>(f * RANK_NB + r);

            if (board[s] == BAN_STONE) {
                const Piece pc = board[s];
                byTypeBB[ALL_PIECES] ^= s;
                byTypeBB[type_of(pc)] ^= s;
                board[s] = NO_PIECE;
                revert_key(s);
            }
        }
    }
}

inline void Position::set_side_to_move(Color c)
{
    sideToMove = c;
    //us = c;
    them = ~sideToMove;
}

inline void Position::change_side_to_move()
{
    set_side_to_move(~sideToMove);
    st.key ^= Zobrist::side;
}

inline Key Position::update_key(Square s)
{
    const int pieceType = color_on(s);

    st.key ^= Zobrist::psq[pieceType][s];

    return st.key;
}

inline Key Position::revert_key(Square s)
{
    return update_key(s);
}

Key Position::update_key_misc()
{
    st.key = st.key << Zobrist::KEY_MISC_BIT >> Zobrist::KEY_MISC_BIT;

    st.key |= static_cast<Key>(pieceToRemoveCount) << (CHAR_BIT * sizeof(Key) - Zobrist::KEY_MISC_BIT);

    return st.key;
}

///////////////////////////////////////////////////////////////////////////////

#include "movegen.h"
#include "misc.h"

Bitboard Position::millTableBB[SQUARE_NB][LD_NB] = {{0}};


void Position::create_mill_table()
{
    Mills::mill_table_init();
}

Color Position::color_on(Square s) const
{
    return color_of(board[s]);
}

bool Position::bitboard_is_ok()
{
#ifdef BITBOARD_DEBUG
    Bitboard whiteBB = byColorBB[WHITE];
    Bitboard blackBB = byColorBB[BLACK];

    for (Square s = SQ_BEGIN; s < SQ_END; ++s) {

        if (empty(s))
        {
            if (whiteBB & (1 << s)) {
                return false;
            }

            if (blackBB & (1 << s)) {
                return false;
            }
        }

        if (color_of(board[s]) == WHITE)
        {
            if ((whiteBB & (1 << s)) == 0) {
                return false;
            }

            if (blackBB & (1 << s)) {
                return false;
            }
        }

        if (color_of(board[s]) == BLACK) {
            if ((blackBB & (1 << s)) == 0) {
                return false;
            }

            if (whiteBB & (1 << s)) {
                return false;
            }
        }
    }
#endif

    return true;
}

int Position::potential_mills_count(Square to, Color c, Square from)
{
    int n = 0;
    Piece locbak = NO_PIECE;

    assert(SQ_0 <= from && from < SQUARE_NB);

    if (c == NOBODY) {
        c = color_on(to);
    }

    if (from != SQ_0 && from >= SQ_BEGIN && from < SQ_END) {
        locbak = board[from];
        board[from] = NO_PIECE;

        CLEAR_BIT(byTypeBB[ALL_PIECES], from);
        CLEAR_BIT(byTypeBB[type_of(locbak)], from);
        CLEAR_BIT(byColorBB[color_of(locbak)], from);
    }

    const Bitboard bc = byColorBB[c];
    const Bitboard *mt = millTableBB[to];

    for (int l = 0; l < LD_NB; l++) {
        if ((bc & mt[l]) == mt[l]) {
            n++;
        }
    }

    if (from != SQ_0 && from >= SQ_BEGIN && from < SQ_END) {
        board[from] = locbak;

        SET_BIT(byTypeBB[ALL_PIECES], from);
        SET_BIT(byTypeBB[type_of(locbak)], from);
        SET_BIT(byColorBB[color_of(locbak)], from);
    }

    return n;
}

int Position::mills_count(Square s)
{
    int n = 0;

    const Bitboard bc = byColorBB[color_on(s)];
    const Bitboard *mt = millTableBB[s];

    for (auto i = 0; i < LD_NB; ++i) {
        if (((bc & mt[i]) == mt[i])) {
            n++;
        }
    }

    return n;
}

bool Position::is_all_in_mills(Color c)
{
    for (Square i = SQ_BEGIN; i < SQ_END; ++i) {
        if (board[i] & ((uint8_t)make_piece(c))) {
            if (!potential_mills_count(i, NOBODY)) {
                return false;
            }
        }
    }

    return true;
}

void Position::surrounded_pieces_count(Square s, int &ourPieceCount, int &theirPieceCount, int &bannedCount, int &emptyCount)
{
    Square moveSquare;

    for (MoveDirection d = MD_BEGIN; d < MD_NB; ++d) {
        moveSquare = static_cast<Square>(MoveList<LEGAL>::adjacentSquares[s][d]);

        if (!moveSquare) {
            continue;
        }

        const enum Piece pieceType = static_cast<Piece>(board[moveSquare]);

        switch (pieceType) {
        case NO_PIECE:
            emptyCount++;
            break;
        case BAN_STONE:
            bannedCount++;
            break;
        default:
            if (color_of(pieceType) == sideToMove) {
                ourPieceCount++;
            } else {
                theirPieceCount++;
            }
            break;
        }
    }
}

bool Position::is_all_surrounded(Color c
#ifdef MUEHLE_NMM
                                 , Square from, Square to
#endif // MUEHLE_NMM
                                ) const
{
    // Full
    if (pieceOnBoardCount[WHITE] + pieceOnBoardCount[BLACK] >= EFFECTIVE_SQUARE_NB)
        return true;

    // Can fly
    if (pieceOnBoardCount[c] <= rule.flyPieceCount &&
        rule.mayFly) {
        return false;
    }

    Bitboard bb = byTypeBB[ALL_PIECES];

#ifdef MUEHLE_NMM
    CLEAR_BIT(bb, from);
    SET_BIT(bb, to);
#endif // MUEHLE_NMM

    for (Square s = SQ_BEGIN; s < SQ_END; ++s) {
        if ((c & color_on(s)) &&
            (bb & MoveList<LEGAL>::adjacentSquaresBB[s]) != MoveList<LEGAL>::adjacentSquaresBB[s]) {
            return false;
        }
    }

    return true;
}

bool Position::is_star_square(Square s)
{
    if (rule.hasDiagonalLines == true) {
        return (s == 17 ||
                s == 19 ||
                s == 21 ||
                s == 23);
    }

    return (s == 16 ||
            s == 18 ||
            s == 20 ||
            s == 22);
}

void Position::print_board()
{
    if (rule.hasDiagonalLines) {
        printf("\n"
                    "31 ----- 24 ----- 25\n"
                    "| \\       |      / |\n"
                    "|  23 -- 16 -- 17  |\n"
                    "|  | \\    |   / |  |\n"
                    "|  |  15-08-09  |  |\n"
                    "30-22-14    10-18-26\n"
                    "|  |  13-12-11  |  |\n"
                    "|  | /    |   \\ |  |\n"
                    "|  21 -- 20 -- 19  |\n"
                    "| /       |      \\ |\n"
                    "29 ----- 28 ----- 27\n"
                    "\n");
    } else {
        printf("\n"
                    "31 ----- 24 ----- 25\n"
                    "|         |        |\n"
                    "|  23 -- 16 -- 17  |\n"
                    "|  |      |     |  |\n"
                    "|  |  15-08-09  |  |\n"
                    "30-22-14    10-18-26\n"
                    "|  |  13-12-11  |  |\n"
                    "|  |      |     |  |\n"
                    "|  21 -- 20 -- 19  |\n"
                    "|         |        |\n"
                    "29 ----- 28 ----- 27\n"
                    "\n");
    }
}

void Position::reset_bb()
{
    memset(byTypeBB, 0, sizeof(byTypeBB));
    memset(byColorBB, 0, sizeof(byColorBB));

    for (Square s = SQ_BEGIN; s < SQ_END; ++s) {
        const Piece pc = board[s];
        byTypeBB[ALL_PIECES] |= byTypeBB[type_of(pc)] |= s;
        byColorBB[color_of(pc)] |= s;
    }
}

void Position::updateMobility(MoveType mt, Square s)
{
    if (!gameOptions.getConsiderMobility()) {
        return;
    }

    Bitboard adjacentWhiteBB = byColorBB[WHITE] & MoveList<LEGAL>::adjacentSquaresBB[s];
    Bitboard adjacentBlackBB = byColorBB[BLACK] & MoveList<LEGAL>::adjacentSquaresBB[s];
    Bitboard adjacentNoColorBB = (~(byColorBB[BLACK] | byColorBB[WHITE])) & MoveList<LEGAL>::adjacentSquaresBB[s];
    int adjacentWhiteBBCount = popcount(adjacentWhiteBB);
    int adjacentBlackBBCount = popcount(adjacentBlackBB);
    int adjacentNoColorBBCount = popcount(adjacentNoColorBB);

    if (mt == MOVETYPE_PLACE) {
        mobilityDiff -= adjacentWhiteBBCount;
        mobilityDiff += adjacentBlackBBCount;

        if (side_to_move() == WHITE) {
            mobilityDiff += adjacentNoColorBBCount;
        } else {
            mobilityDiff -= adjacentNoColorBBCount;
        }
    } else if (mt == MOVETYPE_REMOVE) {
        mobilityDiff += adjacentWhiteBBCount;
        mobilityDiff -= adjacentBlackBBCount;

        if ( color_of(board[s]) == WHITE) {
            mobilityDiff -= adjacentNoColorBBCount;
        } else {
            mobilityDiff += adjacentNoColorBBCount;
        }
    } else {
        assert(0);
    }
}

void Position::mirror(vector <string> &moveHistory, bool cmdChange /*= true*/)
{
    Piece ch;
    int f, r;
    int i;

    for (f = 1; f <= FILE_NB; f++) {
        for (r = 1; r < RANK_NB / 2; r++) {
            ch = board[f * RANK_NB + r];
            board[f * RANK_NB + r] = board[(f + 1) * RANK_NB - r];
            board[(f + 1) * RANK_NB - r] = ch;
        }
    }

    reset_bb();

    uint64_t llp[3] = { 0 };

    if (move < 0) {
        f = (-move) / RANK_NB;
        r = (-move) % RANK_NB;
        r = (RANK_NB - r) % RANK_NB;
        move = static_cast<Move>(-(f * RANK_NB + r));
    } else {
        llp[0] = static_cast<uint64_t>(from_sq((Move)move));
        llp[1] = to_sq((Move)move);

        for (i = 0; i < 2; i++) {
            f = static_cast<int>(llp[i]) / RANK_NB;
            r = static_cast<int>(llp[i]) % RANK_NB;
            r = (RANK_NB - r) % RANK_NB;
            llp[i] = (static_cast<uint64_t>(f) * RANK_NB + r);
        }

        move = static_cast<Move>(((llp[0] << 8) | llp[1]));
    }

    if (currentSquare != 0) {
        f = currentSquare / RANK_NB;
        r = currentSquare % RANK_NB;
        r = (RANK_NB - r) % RANK_NB;
        currentSquare = static_cast<Square>(f * RANK_NB + r);
    }

    if (cmdChange) {
        unsigned r1, s1, r2, s2;
        int args = 0;

        args = sscanf(record, "(%1u,%1u)->(%1u,%1u)", &r1, &s1, &r2, &s2);
        if (args >= 4) {
            s1 = (RANK_NB - s1 + 1) % RANK_NB;
            s2 = (RANK_NB - s2 + 1) % RANK_NB;
            record[3] = '1' + static_cast<char>(s1);
            record[10] = '1' + static_cast<char>(s2);
        } else {
            args = sscanf(record, "-(%1u,%1u)", &r1, &s1);
            if (args >= 2) {
                s1 = (RANK_NB - s1 + 1) % RANK_NB;
                record[4] = '1' + static_cast<char>(s1);
            } else {
                args = sscanf(record, "(%1u,%1u)", &r1, &s1);
                if (args >= 2) {
                    s1 = (RANK_NB - s1 + 1) % RANK_NB;
                    record[3] = '1' + static_cast<char>(s1);
                }
            }
        }

        for (auto &iter : moveHistory) {
            args = sscanf(iter.c_str(), "(%1u,%1u)->(%1u,%1u)", &r1, &s1, &r2, &s2);
            if (args >= 4) {
                s1 = (RANK_NB - s1 + 1) % RANK_NB;
                s2 = (RANK_NB - s2 + 1) % RANK_NB;
                iter[3] = '1' + static_cast<char>(s1);
                iter[10] = '1' + static_cast<char>(s2);
            } else {
                args = sscanf(iter.c_str(), "-(%1u,%1u)", &r1, &s1);
                if (args >= 2) {
                    s1 = (RANK_NB - s1 + 1) % RANK_NB;
                    iter[4] = '1' + static_cast<char>(s1);
                } else {
                    args = sscanf(iter.c_str(), "(%1u,%1u)", &r1, &s1);
                    if (args >= 2) {
                        s1 = (RANK_NB - s1 + 1) % RANK_NB;
                        iter[3] = '1' + static_cast<char>(s1);
                    }
                }
            }
        }
    }
}

void Position::turn(vector <string> &moveHistory, bool cmdChange /*= true*/)
{
    Piece ch;
    int f, r;
    int i;

    for (r = 0; r < RANK_NB; r++) {
        ch = board[RANK_NB + r];
        board[RANK_NB + r] = board[EFFECTIVE_SQUARE_NB + r];
        board[EFFECTIVE_SQUARE_NB + r] = ch;
    }

    reset_bb();

    uint64_t llp[3] = { 0 };

    if (move < 0) {
        f = (-move) / RANK_NB;
        r = (-move) % RANK_NB;

        if (f == 1)
            f = FILE_NB;
        else if (f == FILE_NB)
            f = 1;

        move = static_cast<Move>(-(f * RANK_NB + r));
    } else {
        llp[0] = static_cast<uint64_t>(from_sq((Move)move));
        llp[1] = to_sq((Move)move);

        for (i = 0; i < 2; i++) {
            f = static_cast<int>(llp[i]) / RANK_NB;
            r = static_cast<int>(llp[i]) % RANK_NB;

            if (f == 1)
                f = FILE_NB;
            else if (f == FILE_NB)
                f = 1;

            llp[i] = static_cast<uint64_t>(f * RANK_NB + r);
        }

        move = static_cast<Move>(((llp[0] << 8) | llp[1]));
    }

    if (currentSquare != 0) {
        f = currentSquare / RANK_NB;
        r = currentSquare % RANK_NB;

        if (f == 1)
            f = FILE_NB;
        else if (f == FILE_NB)
            f = 1;

        currentSquare = static_cast<Square>(f * RANK_NB + r);
    }

    if (cmdChange) {
        unsigned r1, s1, r2, s2;
        int args = 0;

        args = sscanf(record, "(%1u,%1u)->(%1u,%1u)",
                      &r1, &s1, &r2, &s2);

        if (args >= 4) {
            if (r1 == 1)
                r1 = FILE_NB;
            else if (r1 == FILE_NB)
                r1 = 1;

            if (r2 == 1)
                r2 = FILE_NB;
            else if (r2 == FILE_NB)
                r2 = 1;

            record[1] = '0' + static_cast<char>(r1);
            record[8] = '0' + static_cast<char>(r2);
        } else {
            args = sscanf(record, "-(%1u,%1u)", &r1, &s1);
            if (args >= 2) {
                if (r1 == 1)
                    r1 = FILE_NB;
                else if (r1 == FILE_NB)
                    r1 = 1;
                record[2] = '0' + static_cast<char>(r1);
            } else {
                args = sscanf(record, "(%1u,%1u)", &r1, &s1);
                if (args >= 2) {
                    if (r1 == 1)
                        r1 = FILE_NB;
                    else if (r1 == FILE_NB)
                        r1 = 1;
                    record[1] = '0' + static_cast<char>(r1);
                }
            }
        }

        for (auto &iter : moveHistory) {
            args = sscanf(iter.c_str(),
                          "(%1u,%1u)->(%1u,%1u)",
                          &r1, &s1, &r2, &s2);

            if (args >= 4) {
                if (r1 == 1)
                    r1 = FILE_NB;
                else if (r1 == FILE_NB)
                    r1 = 1;

                if (r2 == 1)
                    r2 = FILE_NB;
                else if (r2 == FILE_NB)
                    r2 = 1;

                iter[1] = '0' + static_cast<char>(r1);
                iter[8] = '0' + static_cast<char>(r2);
            } else {
                args = sscanf(iter.c_str(), "-(%1u,%1u)", &r1, &s1);
                if (args >= 2) {
                    if (r1 == 1)
                        r1 = FILE_NB;
                    else if (r1 == FILE_NB)
                        r1 = 1;

                    iter[2] = '0' + static_cast<char>(r1);
                } else {
                    args = sscanf(iter.c_str(), "(%1u,%1u)", &r1, &s1);
                    if (args >= 2) {
                        if (r1 == 1)
                            r1 = FILE_NB;
                        else if (r1 == FILE_NB)
                            r1 = 1;

                        iter[1] = '0' + static_cast<char>(r1);
                    }
                }
            }
        }
    }
}

void Position::rotate(vector <string> &moveHistory, int degrees, bool cmdChange /*= true*/)
{
    degrees = degrees % 360;

    if (degrees < 0)
        degrees += 360;

    if (degrees == 0 || degrees % 90)
        return;

    degrees /= 45;

    Piece ch1, ch2;
    int f, r;

    if (degrees == 2) {
        for (f = 1; f <= FILE_NB; f++) {
            ch1 = board[f * RANK_NB];
            ch2 = board[f * RANK_NB + 1];

            for (r = 0; r < RANK_NB - 2; r++) {
                board[f * RANK_NB + r] = board[f * RANK_NB + r + 2];
            }

            board[f * RANK_NB + 6] = ch1;
            board[f * RANK_NB + 7] = ch2;
        }
    } else if (degrees == 6) {
        for (f = 1; f <= FILE_NB; f++) {
            ch1 = board[f * RANK_NB + 7];
            ch2 = board[f * RANK_NB + 6];

            for (r = RANK_NB - 1; r >= 2; r--) {
                board[f * RANK_NB + r] = board[f * RANK_NB + r - 2];
            }

            board[f * RANK_NB + 1] = ch1;
            board[f * RANK_NB] = ch2;
        }
    } else if (degrees == 4) {
        for (f = 1; f <= FILE_NB; f++) {
            for (r = 0; r < RANK_NB / 2; r++) {
                ch1 = board[f * RANK_NB + r];
                board[f * RANK_NB + r] = board[f * RANK_NB + r + 4];
                board[f * RANK_NB + r + 4] = ch1;
            }
        }
    } else {
        return;
    }

    reset_bb();

    uint64_t llp[3] = { 0 };

    if (move < 0) {
        f = (-move) / RANK_NB;
        r = (-move) % RANK_NB;
        r = (r + RANK_NB - degrees) % RANK_NB;
        move = static_cast<Move>(-(f * RANK_NB + r));
    } else {
        llp[0] = static_cast<uint64_t>(from_sq((Move)move));
        llp[1] = to_sq((Move)move);
        f = static_cast<int>(llp[0]) / RANK_NB;
        r = static_cast<int>(llp[0]) % RANK_NB;
        r = (r + RANK_NB - degrees) % RANK_NB;
        llp[0] = static_cast<uint64_t>(f * RANK_NB + r);
        f = static_cast<int>(llp[1]) / RANK_NB;
        r = static_cast<int>(llp[1]) % RANK_NB;
        r = (r + RANK_NB - degrees) % RANK_NB;
        llp[1] = static_cast<uint64_t>(f * RANK_NB + r);
        move = static_cast<Move>(((llp[0] << 8) | llp[1]));
    }

    if (currentSquare != 0) {
        f = currentSquare / RANK_NB;
        r = currentSquare % RANK_NB;
        r = (r + RANK_NB - degrees) % RANK_NB;
        currentSquare = static_cast<Square>(f * RANK_NB + r);
    }

    if (cmdChange) {
        unsigned r1, s1, r2, s2;
        int args = 0;

        args = sscanf(record, "(%1u,%1u)->(%1u,%1u)", &r1, &s1, &r2, &s2);
        if (args >= 4) {
            s1 = (s1 - 1 + RANK_NB - degrees) % RANK_NB;
            s2 = (s2 - 1 + RANK_NB - degrees) % RANK_NB;
            record[3] = '1' + static_cast<char>(s1);
            record[10] = '1' + static_cast<char>(s2);
        } else {
            args = sscanf(record, "-(%1u,%1u)", &r1, &s1);

            if (args >= 2) {
                s1 = (s1 - 1 + RANK_NB - degrees) % RANK_NB;
                record[4] = '1' + static_cast<char>(s1);
            } else {
                args = sscanf(record, "(%1u,%1u)", &r1, &s1);

                if (args >= 2) {
                    s1 = (s1 - 1 + RANK_NB - degrees) % RANK_NB;
                    record[3] = '1' + static_cast<char>(s1);
                }
            }
        }

        for (auto &iter : moveHistory) {
            args = sscanf(iter.c_str(), "(%1u,%1u)->(%1u,%1u)", &r1, &s1, &r2, &s2);

            if (args >= 4) {
                s1 = (s1 - 1 + RANK_NB - degrees) % RANK_NB;
                s2 = (s2 - 1 + RANK_NB - degrees) % RANK_NB;
                iter[3] = '1' + static_cast<char>(s1);
                iter[10] = '1' + static_cast<char>(s2);
            } else {
                args = sscanf(iter.c_str(), "-(%1u,%1u)", &r1, &s1);

                if (args >= 2) {
                    s1 = (s1 - 1 + RANK_NB - degrees) % RANK_NB;
                    iter[4] = '1' + static_cast<char>(s1);
                } else {
                    args = sscanf(iter.c_str(), "(%1u,%1u)", &r1, &s1);
                    if (args >= 2) {
                        s1 = (s1 - 1 + RANK_NB - degrees) % RANK_NB;
                        iter[3] = '1' + static_cast<char>(s1);
                    }
                }
            }
        }
    }
}
